C*----------------------------------------------------------------------
C*
C*  NAME:
C*    pfrdp, pfrda, pfrdb, pfrdc - Seismic Processing Benchmark 
C*        process PFRD
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call pfrdp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call pfrda( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call pfrdb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call pfrdc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Single threaded read from Seismic Benchmark Format dataset
C*
C*  PARAMETERS: 
C*    
C*    Name      Description                           Default
C*    path      Path name of PBF format file          Required
C*    mode      Read Mode:                            Block
C*                RoundRob - Round robin
C*                Block    - Block assignment 
C*    nfiles    Number of .TRC files to split         Optional
C*              dataset into.  Provided for
C*              striping and cluster suport.
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./group.inc
C*       ./para.inc
C*       ./pfrd.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     pfrdp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process pfrds 'PFRD' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine pfrdp( ldim, maxtrc, nra, nsa, abort, ipr )
	implicit none
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C
C  Common blocks:
C
	include 'group.inc'
	include 'para.inc'
	include 'pfrd.inc'

C
C  Local declarations:
C
	character*80 buf
	integer jtemp, jfiles, ier
	integer sbfopen, sbfclose, sbfqfiles

C
C  Start of executable statements
C

C  Get parameter record with file name
	call sysparm( 'path', 'C', dsn, ier )
	if (ier.ne.0) then
		write (ipr,*) 'Parameter "path" required for process PFRD'
		abort = .true.
		return
	endif

C  Debug flag
	idbug = 0
	call sysparm( 'debug', 'C', buf, ier )
	if (ier.eq.0 .and. buf(1:3).eq.'yes') idbug = 1

C  Read mode
	buf = 'Block'
	call sysparm( 'mode', 'C', buf, ier )
	if (buf(1:5).eq.'Panel') then
C  Panel decomposition - one group spread across nodes
		jmode = JPAN
	else if (buf(1:5).eq.'Trans') then
C  Transposed read - round-robin assignment of groups across nodes
		jmode = JTRN
	else
C  Normal read - block assignment of groups across nodes
		jmode = JBLK
	endif

C  Serialized read
	rdpar = .true.
	call sysparm('serial', 'C', buf, ier )
	if (ier.eq.0 .and. buf(1:4).eq.'true') rdpar = .false.

C  Attempt to open
	handle = sbfopen( dsn, 'r ' )
	if (handle .le. 0) then
		write (ipr,*) 'PFRD: could not open ',dsn
		call sbferror()
		abort = .true.
		return
	endif
	
C  Get file info
	call sbfinfo( handle, jns, jntpg, jngpl, jnlin )
	call sbfrinfo( handle, jmsr, rx0, ry0, rdx, rdy, rrot, rtype )
	jfiles = sbfqfiles( handle )

C  Switch TrcPrGrp and GrpPrLin if in transpose mode
	if (jmode.eq.JTRN) then
		jtemp = jntpg
		jntpg = jngpl
		jngpl = jtemp
		rtype = 'xoff    '
	endif

C  Set other memory requirements
	ldim = jns
	if (jmode.eq.JPAN) then
		call jdcomp( jngpl, jntpn, jnlive )
		maxtrc = jntpn
	else
		maxtrc = jntpg
	endif
	nra = nodes
	nsa = 6*maxtrc

C  Write out info
	write (ipr,9010) dsn,rtype,jns,jntpg,jngpl,jnlin,jfiles,jmsr,
     +    rx0,ry0,rdx,rdy,rrot
 9010 format (/,' Process PFRD: Single Threaded Benchmark File Read',/,
     +          ' Path name        : ',a64,/,
     +          ' Group type       : ',4x,a8,/,
     +          ' Samples per Trace: ',i8,/,
     +          ' Traces per Group : ',i8,/,
     +          ' Groups per Line  : ',i8,/,
     +          ' Number of Lines  : ',i8,/,
     +          ' Number of Files  : ',i8,/,
     +          ' Sample Rate (ms) : ',i8,/,
     +          ' CDP Grid Orgin   : ',2f10.1,/,
     +          ' CDP Separation   : ',f10.1,/,
     +          ' Line Separation  : ',f10.1,/,
     +          ' Grid Rotation    : ',f10.1)

	if (jmode.eq.JTRN) then
		write (ipr,*) ' Data will be read in transposed mode'
		write (ipr,'('' Input Size   : '',4i6)') 
     +      jns,jngpl,jntpg,jnlin
	else if (jmode.eq.JPAN) then
		write (ipr,*) ' Data will be read in paneled mode'
	else if (jmode.eq.JBLK) then
		write (ipr,*) ' Data will be read in block mode'
	endif

	if (.not. rdpar) 
     +	write (ipr,*) ' Read will be serialized on master node'

	if (idbug.ne.0) write (ipr,*) 'Debug print is on'

C  Close file
	ier = sbfclose( handle )

C  Set attributes for subsequent processes
	call pfrdatr()

	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     pfrda, pfrdb, pfrdc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for reading Parallel Benchmark 
C*     Format files
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        sbfread - read group from SBF file
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine pfrda( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	implicit none
C  Argument declartations
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'para.inc'
	include 'pfrd.inc'
C  Local decalarations
	integer sbfopen, jmsgrng

C  Debug print
	if (idbug.ne.0) idbug = ipr

c  Initialize init flag
	init = .true.

c  Open data set
	if (rdpar .or. node.eq.master) then
		handle = sbfopen( dsn, 'r ' )
		if (handle .le. 0) then
			write (ipr,*) 'PFRD: open failed on node ', node
			call sbferror()
			abort = .true.
			return
		endif
	endif

C  Set magic number in common 
	magic = 12345

C  Set trace decomposition and range for each mode
	if (jmode.eq.JPAN) then
C  Panel decomposition read
		call jdcomp( jngpl, jntpn, jnlive )
		jgrp1 = 1
		jgrpn = jntpg
		jgrpi = 1
	else
C  Transposed and Block read
		call jdcomp( jngpl, jntpn, jnlive )
C  Set group range on each node
		call jrange( jngpl, jgrp1, jgrpn )
		if (jgrp1.eq.0) jgrp1 = 1
		jgrpi = 1
	endif
	if (idbug.ne.0) write (ipr,9010) node,jntpn,jnlive
 9010 format (' PFRDA: node, jntpn, jnlive = ',3i5)

c  Set attributes
	call pfrdatr()

C  Intialize I/O timer
	etio = 0.0

C  Intialize done flags for serial mode
	if (.not. rdpar) then
		call vfill( 0, ra, 1, nodes )
		msgbase = jmsgrng( nodes )
	endif

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine does nothing
C*----------------------------------------------------------------------

	subroutine pfrdb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	ntro = 0
	return
	end

C*----------------------------------------------------------------------
C*  All the action is in the 'c' routine
C*----------------------------------------------------------------------
c
	subroutine pfrdc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	integer ra(nra)
	real otr(ldim,maxtrc), sa(nsa)
	logical abort
C  Common blocks
	include 'group.inc'
	include 'coord.inc'
	include 'para.inc'
	include 'pfrd.inc'
C  Local declarations
	integer i,j,k
	real bytesm,rate

C  First time in, set initial values for group and line numbers
	if (init) then
		jgroup = jgrp1 - jgrpi
		jline = 1
		init = .false.
		call jsync()
	endif

c  Increment group counter
	jgroup = jgroup + jgrpi

c  check for line increment
	if (jgroup.gt.jgrpn) then
		jline = jline + 1
C  Check for end of data set
		if (jline.gt.jnlin) then
			ntro = 0
C  Read is complete, send termination to master node
			call pfrd_mclean( ra )
			if (node.ne.master) return
C  I/O total (Mbytes) and rate (MB/s)
			bytesm = 4.e-6 * float(jntpg*jngpl*jnlin)*float(jns+6)
			if (etio.eq.0.) etio = 1.
			rate = bytesm/etio
			write (ipr,9010) bytesm, etio, rate
 9010 format (/,' Process PFRD Completed',/,
     +          ' Mbytes read     : ',f10.3,/,
     +          ' I/O Elapsed Time: ',f10.3,/,
     +          ' I/O Rate (MB/s) : ',f10.3,/)
			return
		endif
C  New line, reset group counter
		jgroup = jgrp1
c  Sync if in block mode so each line starts at same time
		if (jmode.eq.JBLK) call jsync()
	endif

C  Set attributes
	call pfrdatr()
	ntro = ntpg

c  Read next group
	call pfrd_rdnext( ldim, maxtrc, otr, sa, ra, abort, ipr )

C  Set coordinates
	k = 1
	do j=1,ntro
		do i=1,3
			sxyz(i,j) = sa(k)
			k = k + 1
		enddo
		do i=1,3
			rxyz(i,j) = sa(k)
			k = k + 1
		enddo
	enddo

	return
	end

C*----------------------------------------------------------------------

	subroutine pfrdatr()
C
C  Set attributes in system commons
C
	implicit none
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'pfrd.inc'

C  Set data set attributes
	ns = jns
	msr = jmsr
	ntpg = jntpg
	ngpl = jngpl
	nlin = jnlin
	gtype = rtype
	gx0 = rx0
	gy0 = ry0
	gdx = rdx
	gdy = rdy
	grot = rrot
	igroup = jgroup
	iline = jline

C  Parallel decomposition attributes
	if (jmode.eq.JPAN) then
		pmode = 1
		ntpn = jntpn
		nlive = jnlive
	else
		pmode = 0
		ntpn = ntpg
		nlive = ntpg
	endif

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine pfrd_attr( name, attr, len )
C
C  Return input data stream attribute:
C  name - input attribute name
C         This is a character string that defines the attribute type.
C         i.e. 'zaxis': depth locations for variable sampling
C  attr - output attribute array, i.e. output array of depth locations
C  len - output attribute length in bytes, -1 if attribute is not there
C        To get length in words, use lenwords = len / jwsize()
C
	implicit none
	character*(*) name
	real attr(*)
	integer len, sbfrdatr
	include 'pfrd.inc'

C  Return error if data set is not open
	len = -1
	if (magic.ne.12345) return

C  Read attribute
	len = sbfrdatr( handle, name, attr )

	return
	end

C
C-----------------------------------------------------------------------
C

	subroutine pfrd_rdnext( ldim, maxtrc, otr, sa, ra, abort, ipr )
C 
C  Read next frame from SBF dataset.
C  Handles both parallel and serialized reads
C
C  otr(ldim,maxtrc) - output next record
C  sa(6,maxtrc) - output coordinates
C  ra(nodes) - node termination flag array
C  abort - output abort flag
C
	implicit none
C  Arguments
	integer ldim, maxtrc, ipr
	real otr(ldim,maxtrc), sa(*)
	integer ra(*)
	logical abort
C  Local variables
	integer ier, msg, inode, len, ibuf(2)
	integer sbfread, sbfreadp, sbfreadt, jwsize
	real t0, etimer
	include 'para.inc'
	include 'pfrd.inc'

C  Read data and return if in parallel mode
	if (rdpar) then
		if (idbug.ne.0) write (ipr,9010) node,jgroup,jline
 9010 	format (' PFRDC: node, group, line = ',3i5)
		ier = 0
		if (jmode.eq.JTRN) then
			ier = sbfreadt( handle, jgroup, jline, ldim, maxtrc, 
     +			otr, sa )
		else if (jmode.eq.JPAN) then
			ier = sbfreadp( handle, jgroup, jline, ldim, maxtrc,
     +			otr, sa )
		else
			ier = sbfread( handle, jgroup, jline, ldim, maxtrc,
     +			otr, sa )
		endif
C  Process errors
		if (ier.ne.0) then
			write (ipr,9020) jgroup,jline,ier
 9020   	format (' PFRD ERROR - group ',i5,' line ',i5,' ier ',i5)
			call sbferror()
			abort = .true.
			return
		endif
		return
	endif

C  Serialized read on master node

C  If not master node, send request to master and then receive
	if (node.ne.master) then
C  Send request for group number
		len = 2*jwsize()
		ibuf(1) = jgroup
		ibuf(2) = jline
		msg = msgbase + node
		call jsend( msg, ibuf, len, master )
C  Receive coords and traces
		len = 6*maxtrc*jwsize()
		call jrecv( msg, sa, len )
		len = ldim*maxtrc*jwsize()
		call jrecv( msg, otr, len )
		return
	endif

C  Master node loops and processes requests
	t0 = etimer(0)
	do 10 inode = 1, nodes-1
C  Check for this node done reading
		if (ra(inode).ne.0) goto 10
C  Recive node numbers
		len = 2*jwsize()
		msg = msgbase + inode
		call jrecv( msg, ibuf, len )
C  Check for terminate message
		if (ibuf(1).le.0) then
			ra(inode) = -1
			goto 10
		endif

		if (idbug.ne.0) write (ipr,9010) inode,ibuf(1),ibuf(2)

C  Read requested data
		ier = 0
		if (jmode.eq.JTRN) then
			ier = sbfreadt(handle, ibuf(1), ibuf(2), 
     +			ldim, maxtrc, otr, sa )
		else if (jmode.eq.JPAN) then
			ier = sbfreadp(handle, ibuf(1), ibuf(2), 
     +			ldim, maxtrc, otr, sa )
		else
			ier = sbfread(handle, ibuf(1), ibuf(2), 
     +			ldim, maxtrc, otr, sa )
		endif
		if (ier.ne.0) then
			write (ipr,9020) ibuf(1),ibuf(2),ier
			call sbferror()
			abort = .true.
			return
		endif

C  Send data back to requesting node
		len = 6*maxtrc*jwsize()
		call jsend( msg, sa, len, inode )
		len = ldim*maxtrc*jwsize()
		call jsend( msg, otr, len, inode )

  10  continue

C  Read data for master node
	if (idbug.ne.0) write (ipr,9010) node,jgroup,jline
	ier = 0
	if (jmode.eq.JTRN) then
		ier = sbfreadt( handle, jgroup, jline, ldim, maxtrc, otr, sa )
	else if (jmode.eq.JPAN) then
		ier = sbfreadp( handle, jgroup, jline, ldim, maxtrc, otr, sa )
	else
		ier = sbfread( handle, jgroup, jline, ldim, maxtrc, otr, sa )
	endif

C  Update I/O time
	etio = etio + etimer(1)

C  Process errors
	if (ier.ne.0) then
		write (ipr,9020) jgroup,jline,ier
		call sbferror()
		abort = .true.
		return
	endif

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine pfrd_mclean( ra )
C
C  Nodes send termination message to master
C  Master node receives any excess termination messages
C  
	implicit none
	integer ra(*)
	integer len, inode, ibuf(2), jwsize
	include 'para.inc'
	include 'pfrd.inc'

C  If in parallel mode, close file and return
	if (rdpar) then
		call sbfclose(handle)
		return
	endif

	len = 2*jwsize()

C  Send termination message to master
	if (node.ne.master) then
		ibuf(1) = -1
		ibuf(2) = 0
		call jsend( msgbase + node, ibuf, len, master )
		return
	endif

C  Loop over nodes
	do inode = 1, nodes-1
C  Receive outstanding termination message if needed
		if (ra(inode).eq.0) then
			call jrecv( msgbase + inode, ibuf, len )
			ra(inode) = -1
		endif
	enddo

C  Close file
	call sbfclose(handle)

	return
	end
C
C-----------------------------------------------------------------------
