C*----------------------------------------------------------------------
C*
C*  NAME:
C*    seis.f - driver program for parallel Seismic Benchmarks
C*     
C*  SYNOPSIS:
C*    call seisprep( )
C*    call seisproc( otr, ra, sa )
C*
C*  DESCRIPTION: 
C*    Routine 'seisprep' executes the prep phase and returns memory
C*    requirements.  Routine 'seisproc' manages the proc phase.
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     sys.inc - seismic benchmark system common
C*     sysprep.f - preparatory subroutine handler
C*     sysproc.f - processing subroutine handler
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*     Written:
C*       12/31/91 - C. Mosher, ARCO E&PT 
C*
C*     References:
C*       Seismic Benchmark Users Guide
C*
C*     Modified:
C*
C*----------------------------------------------------------------------
C
C*----------------------------------------------------------------------
C*  NAME:
C*    seisprep - manage control flow for prep phase
C*     
C*----------------------------------------------------------------------
C*  SYNOPSIS:
C*
C*    call seisprep( )
C*
C*----------------------------------------------------------------------

	subroutine seisprep( )
C
C  Run prep phase and return memory requirements
C  lotr - Output length of trace array in words
C  lra - output length of reserved array in words
C  lsa - output length of scratch array in words
C
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'sys.inc'
	logical abort
	character*80 buf
	integer ier, uprep
 
C*----------------------------------------------------------------------

c  Initialize elapsed timer
	et0 = etimer(0)

C  Master node gets processing flow
	if (node.eq.master) then
		call sysohdr('SEIS')

c  Parse 'PROC' records for process names
		call syspars( )
		if (nproc.eq.0) then
			write (ipr,*) '*** SEIS ERROR: invalid PROC record'
			call sysexit('*** SEIS terminating ...')
		endif

c  Print processing flow
		call syspflo( )
	endif

c  Broadcast process name list
	call sysbcst( sys_start, sys_end )

c  Call parameter check routine for each process in flow
c  Check for abort
c  Update system common block

	ldm = 0
	mxtrc = 0
	lra = 0
	lsa = 0
	abort = .false.
	do jproc = 1, NPROC

		call sysohdr(name(jproc))
		ct = timer(0)
		et = etimer(0)
		ier = uprep( name(jproc), ldm, mxtrc, lra, lsa, abort, ipr )
		if (ier.ne.0) call sysprep(  
     +      name(jproc), ldm, mxtrc, lra, lsa, abort, ipr )
		etime(jproc) = etime(jproc) + etimer(0) - et
		ctime(jproc) = ctime(jproc) + timer(0) - ct
		if (node.eq.master) then
			if (abort) call sysabrt( )
			call sysupd( ldm, mxtrc )
		endif

	enddo

c  Master node prints out memory requirements
	if (node.eq.master) call syschk( )

c  Broadcast system commons
	call sysbcst( sys_start, sys_end )
	call sysbcst( coord_start, coord_end )
	call sysbcst( group_start, group_end )

c  and return
	return
	end


C*----------------------------------------------------------------------
C*  NAME:
C*    seisproc - manage data and control flow for proc phase
C*     
C*----------------------------------------------------------------------
C*  SYNOPSIS:
C*
C*    real otr(*),ra(*),sa(*)
C*    call seisproc( otr, ra, sa )
C*
C*----------------------------------------------------------------------

	subroutine seisproc( otr, ra, sa )
	real otr(*), ra(*), sa(*)
	include 'sys.inc'
	include 'para.inc'
	logical abort
	integer ier, uproc

C*----------------------------------------------------------------------

C  Print proc step heading
	call sysohdr('PROC')

c  Call initialization routine for each process
	
	abort = .false.
	do jproc = 1, nproc
		etime(jproc) = 0.0
		ctime(jproc) = 0.0
		ct = timer(0)
		et = etimer(0)
		ier = uproc( name(jproc), 'A',
     +		ldim, maxtrc, otr, lenra(jproc), ra(kra(jproc)), 
     +		nsa, sa, ntri, ntro, abort, ipr )
		if (ier.ne.0) call sysproc( name(jproc), 'A',
     +		ldim, maxtrc, otr, lenra(jproc), ra(kra(jproc)), 
     +		nsa, sa, ntri, ntro, abort, ipr )
		etime(jproc) = etime(jproc) + etimer(0) - et
		ctime(jproc) = ctime(jproc) + timer(0) - ct
		if (abort) call sysabrt( )
	enddo

C*----------------------------------------------------------------------

c  Call processing routines for each process

c  Initialize number of input and output traces to zero
	ntri = 0
	ntro = 0
	abort = .false.

c  jtop is a pointer to the current process originating traces
	jtop = 1

c  Main processing loop
 100  continue

c  If last process is done originating traces, then quit
	if (jtop.gt.nproc) then
		call sysohdr('SEIS')
		call sysdone( )
		return
	endif

c  For each process in flow, call output routine if process is
c  currently the originating process, otherwise call processing routine

 	do jproc = jtop, nproc

		if (jtop.eq.jproc) then
			ct = timer(0)
			et = etimer(0)
			ier = uproc( name(jproc), 'C',
     +    		ldim, maxtrc, otr, lenra(jproc), ra(kra(jproc)), 
     +			nsa, sa, ntri, ntro, abort, ipr )
			if (ier.ne.0) call sysproc( name(jproc), 'C',
     +    		ldim, maxtrc, otr, lenra(jproc), ra(kra(jproc)), 
     +			nsa, sa, ntri, ntro, abort, ipr )
c  Accumulate timing info and check for abort
			etime(jproc) = etime(jproc) + etimer(0) - et
			ctime(jproc) = ctime(jproc) + timer(0) - ct
			if (abort) call sysabrt( )
c  If zero output traces, increment originating processes
			if (ntro.eq.0) then
				jtop = jtop + 1
				goto 100
			endif
		else
			ntri = ntro
			ct = timer(0)
			et = etimer(0)
			ier = uproc( name(jproc), 'B',
     +    		ldim, maxtrc, otr, lenra(jproc), ra(kra(jproc)), 
     +			nsa, sa, ntri, ntro, abort, ipr )
			if (ier.ne.0) call sysproc( name(jproc), 'B',
     +    		ldim, maxtrc, otr, lenra(jproc), ra(kra(jproc)), 
     +			nsa, sa, ntri, ntro, abort, ipr )
			etime(jproc) = etime(jproc) + etimer(0) - et
			ctime(jproc) = ctime(jproc) + timer(0) - ct
			if (abort) call sysabrt( )
			if (ntro.eq.0) goto 100
		endif

	enddo

	goto 100

	end
