/*--------------------------------------------------------------------*/
/*  NAME:
/*    message.c - message passing routines for ARCO/Perfect benchmark
/*    Version for PVM release 2.4
/*
/*  Routines defined in this module:
/*    jpinit - parallel initialization
/*    jmaster - return master node number
/*    jnodes - return number of nodes in system
/*    jdim - return hypercube dimension
/*    jnode - return node number
/*    jsync - global sync
/*    jkill - kill all processes
/*    jsend - send message (synchronous)
/*    jrecv - receive message (synchronous)
/*    jasend - send message (async)
/*    jarecv - receive message (async)
/*    jamsgwt - async message wait
/*    jbrdcst - broadcast message 
/*    jtrace - toggle tracing
/*    jdcomp - simple decomposition
/*    jrange - return global decomposition index range
/*    jtran23- transpose primitive
/*     
/*--------------------------------------------------------------------*/
/*
/*  HISTORY:
/*
/*     Written:
/*       01/19/92 - C. Mosher, ARCO E&PT 
/*
/*     References:
/*       Intel iPSC/860 System Documentation
/*
/*     Modified:
/*
/*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*/
/*  Global variables
/*--------------------------------------------------------------------*/

/* Trace flag */
static int trace = 0;

/* Node numbers, master node */
static int node, nodes, master;

/* Data decomposition parameters */
static int nglob=0, ntpn=0, nlive=0;

#ifdef PVM
/* local copy of instance name */
static char pvmname[32];
/* internal counter for creating unique PVM barrier names */
static int barnum = 0;
#endif

#ifdef cray
#define ISCRAY
#endif
#ifdef CRAY
#define ISCRAY
#endif

#ifdef ISCRAY
#define jpinit_ JPINIT
#define jpterm_ JPTERM
#define jmaster_ JMASTER
#define jnodes_ JNODES
#define jnode_ JNODE
#define jkill_ JKILL
#define jsync_ JSYNC
#define jsend_ JSEND
#define jrecv_ JRECV
#define jasend_ JASEND
#define jarecv_ JARECV
#define jamsgwt_ JAMSGWT
#define jbrdcst_ JBRDCST
#define jtrace_ JTRACE
#define jsystem_ JSYSTEM
#define jdcomp_ JDCOMP
#define jrange_ JRANGE
#define jtran23_ JTRAN23
#endif

/*--------------------------------------------------------------------*/
/*  Include files
/*--------------------------------------------------------------------*/

#ifdef PVM
#include <stdio.h>
#include "pvmuser.h"
#endif

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jpinit - Initialize parallel processing
/*     
/*  SYNOPSIS:
/*
/*    char *path;
/*    int iret;
/*    iret = jpinit( path );
/*
/*  Initiates parallel processing environment.  For machines that
/*  require host or node initialization, the 'path' variable can
/*  be used to specify the executable to be loaded.
/*
/*  PVM Notes:
/*  Assumes standard PVM installation, and that 'pvmd' has been
/*  started.  Uses the 'pvm_mstat' routine to find out how many
/*  machines are in the cluster.  The number of nodes started is
/*  set to the number of machines in the cluster.  If evironment
/*  variable 'PVM_NODES' is set, this number is used instead.
/*
/*--------------------------------------------------------------------*/

int jpinit( path )
char *path;
{
#ifdef PVM
	int iret,i,enodes;
	char *p,*getenv();
#endif

	nodes = 1;
	node = 0;
	master = 0;

#ifdef PVM
/* Enroll node program */
	iret = enroll( path );
	if (iret < 0) {
		fprintf(stderr,"SeisPerf/PVM: Could not enroll %s\n",path);
		pvm_perror("jpinit");
		return( -1 );
	}

/* Get node instance number */
	iret = whoami( pvmname, &node );
	if (iret < 0) {
		fprintf(stderr,"SeisPerf/PVM: Initialization failed\n");
		pvm_perror("jpinit: whoami");
		return( -1 );
	}

/* Master node spawns additional processes */
	if (node == master) {
/* Get node stats */
		iret = pvm_mstat( &nodes, 0, 0, 0 );
		if (iret < 0) {
			fprintf(stderr,"SeisPerf/PVM: Initialization failed\n");
			pvm_perror("jpinit: pvm_mstat");
			return( -1 );
		}
/* Check for environemnt override of number of nodes */
		if ( (p=getenv("PVM_NTASK")) != (char *)NULL) {
			enodes = 0;
			sscanf(p,"%d",&enodes);
			if (enodes < 1 || enodes > 1024) {
				fprintf(stderr,"SeisPerf/PVM: Initialization failed\n");
				fprintf(stderr,"jpinit: invalid PVM_NTASK value\n");
				return(-1);
			}
			nodes = enodes;
		}
/* Spawn node processes */
		for (i=1; i<nodes; i++) {
			iret = initiate( path, 0 );
			if (iret < 0) {
				fprintf(stderr,"SeisPerf/PVM: Initialization failed\n");
				pvm_perror("jpinit");
				return( -1 );
			}
		}
/* Broadcast number of nodes in the cluster */
		jbrdcst( 1, &nodes, sizeof(int), master );
		printf("SeisPerf/PVM Intialized: %s,  %d nodes\n",
			pvmname,nodes);
	}
	else {
/* Other nodes receive number of nodes */
		jbrdcst( 1, &nodes, sizeof(int), master );
	}
#endif

	return( 0 );
}

#ifndef ISCRAY
int jpinit_( path )
char *path;
{
	return( jpinit( path ) );
}
#endif
#ifdef ISCRAY
#include "fortran.h"
int jpinit_( path )
_fcd path;
{
	char *p, *malloc();
	int len;

	len = _fcdlen(path);
	p = malloc( len + 1 );
	strncpy( p, _fcdtocp(path), len );
	p[len] = '\0';

	return( jpinit( p ) );
}
#endif

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jpterm - terminate parallel processing
/*     
/*  SYNOPSIS:
/*
/*    ier = jpterm();
/*
/*--------------------------------------------------------------------*/

int jpterm()
{

#ifdef PVM
	leave();
#endif

	return( 0 );
}

int jpterm_()
{
	return( jpterm() );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jmaster - return master node number
/*     
/*  SYNOPSIS:
/*
/*    integer master
/*    master = jmaster( )
/*
/*--------------------------------------------------------------------*/

int jmaster()
{
	if (trace)
		printf("jmaster node %d\n",node);
	return(0);
}

int jmaster_()
{
	return( jmaster() );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jnodes - return number of nodes in message passing complex
/*     
/*  SYNOPSIS:
/*
/*    integer nodes
/*    nodes = jnodes( )
/*
/*--------------------------------------------------------------------*/

int jnodes()
{
	if (trace)
		printf("jnodes  node %d\n",node);
	return( nodes );
}

int jnodes_()
{
	return( nodes );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jnode - return processor node number
/*     
/*  SYNOPSIS:
/*
/*    integer node
/*    node = jnode( )
/*
/*--------------------------------------------------------------------*/

int jnode()
{
	if (trace)
		printf("jnode   node %d\n",node);
	return( node );
}

int jnode_()
{
	return( node );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jkill - issue kill signal to entire processor complex
/*     
/*  SYNOPSIS:
/*
/*    call jkill( )
/*
/*--------------------------------------------------------------------*/

int jkill()
{
	int i,iret;

	iret = 0;
	if (trace)
		printf("jkill   node %d\n",node);

#ifdef PVM
	for (i=0; i<nodes; i++)
		if (node != i) iret = terminate( pvmname, i );
#endif

	return( iret );
}

int jkill_()
{
	return( jkill() );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jsync - issue global sync signal to entire processor complex
/*     
/*  SYNOPSIS:
/*
/*    call jsync( )
/*
/*--------------------------------------------------------------------*/

int jsync()
{
#ifdef PVM
	char barnam[12];
#endif

	if (trace)
		printf("jsync   node %d\n",node);

#ifdef PVM
/* Generate unique name, then increment counter */
/*
	sprintf(barnam,"jsync%d",barnum);
	barnum++;
	barrier( barnam, nodes );
*/
	jbrdcst( 99, barnam, 1, master );
#endif

	return( 0 );
}

int jsync_()
{
	return( jsync() );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jsend - send message
/*     
/*  SYNOPSIS:
/*
/*    integer type, length, dest
/*    character*1 buf(length)
/*    call jsend( type, buf, length, dest )
/*
/*--------------------------------------------------------------------*/

int jsend( type, buf, length, dest )
int type, length, dest;
char *buf;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jsend   node %d: type, buf, length, dest = %d 0x%x %d %d\n",
			node,type,buf,length,dest);

#ifdef PVM
/* Select raw data mode */
	pvm_inithow( PvmDataRaw );
/* Copy message into buffer */
	if (length) putbytes( buf, length );
/* Send it on out */
	iret = snd( pvmname, dest, type );
#endif

	return( iret );
}

int jsend_( type, buf, length, node )
int *type, *length, *node;
char *buf;
{
	return( jsend( *type, buf, *length, *node ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jrecv - receive next message
/*     
/*  SYNOPSIS:
/*
/*    integer type, length
/*    character*1 buf(length)
/*    call jrecv( type, buf, length )
/*
/*--------------------------------------------------------------------*/

int jrecv( type, buf, length )
int type, length;
char *buf;
{
	int len,iret;

	iret = 0;
	if (trace)
		printf("jrecv   node %d: type, buf, length = %d 0x%x %d\n",
			node,type,buf,length);

#ifdef PVM
/* wait for message */
	rcv( type );
/* Get length */
	rcvinfo( &len, 0, 0, 0 );
/* Unpack data */
	if (len) iret = getbytes( buf, length );
#endif

	return( iret );
}

int jrecv_( type, buf, length )
int *type, *length;
char *buf;
{
	return( jrecv( *type, buf, *length ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jasend - asynchronous send message
/*     
/*  SYNOPSIS:
/*
/*    integer msg, type, length, dest
/*    character*1 buf(length)
/*    msg = jasend( type, buf, length, dest )
/*
/*--------------------------------------------------------------------*/

int jasend( type, buf, length, dest )
int type, length, dest;
char *buf;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jasend   node %d: type, buf, length, dest = %d 0x%x %d %d\n",
			node,type,buf,length,dest);

#ifdef PVM
	jsend( type, buf, length, dest );
	iret = type;
#endif

	return( iret );
}

int jasend_( type, buf, length, node )
int *type, *length, *node;
char *buf;
{
	return( jasend( *type, buf, *length, *node ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jarecv - asynchronous message receive
/*     
/*  SYNOPSIS:
/*
/*    integer msg, type, node, length
/*    character*1 buf(length)
/*    msg = jarecv( type, buf, length, node )
/*
/*--------------------------------------------------------------------*/

int jarecv( type, buf, length )
int type, length;
char *buf;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jarecv   node %d: type, buf, length = %d 0x%x %d\n",
			node,type,buf,length);

#ifdef PVM
	iret = jrecv( type, buf, length );
#endif

	return( iret );
}

int jarecv_( type, buf, length )
int *type, *length;
char *buf;
{
	return( jarecv( *type, buf, *length ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jamsgwt - asynchronous message receive
/*     
/*  SYNOPSIS:
/*
/*    integer id
/*    call jamsgwt( id )
/*
/*--------------------------------------------------------------------*/

int jamsgwt( id )
int id;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jamsgwt   node %d: id = %d\n", node, id);

#ifdef PVM
/* Loop until message is available */
	while ( (iret = probe(id)) == PvmNoMsg)
		;
	if (iret > 0) iret = 0;
#endif

	return( iret );
}

int jamsgwt_( id )
int *id;
{
	return( jamsgwt( *id ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jbrdcst - broadcast message to all nodes
/*     
/*  SYNOPSIS:
/*
/*    integer type, length, sender
/*    character*1 buf(length)
/*    call jbrdcst( type, buf, length, sender )
/*
/*--------------------------------------------------------------------*/

int jbrdcst( type, buf, length, sender )
int type, length, sender;
char *buf;
{
	int iret;

	iret = 0;
	if (trace)
		printf("jbrdcst node %d: type, buf, length, sender = %d 0x%x %d %d\n",
			node,type,buf,length,sender);

#ifdef PVM
	if (node == sender)
		iret = jsend( type, buf, length, -1 );
	else
		iret = jrecv( type, buf, length );
#endif

	return( iret );
}

jbrdcst_( type, buf, length, sender )
int *type, *length, *sender;
char *buf;
{
	return( jbrdcst( *type, buf, *length, *sender ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jtrace - Toggle trace option
/*     
/*  SYNOPSIS:
/*
/*    integer iop
/*    call jtrace( iop )
/*
/*--------------------------------------------------------------------*/

int jtrace( iop )
int iop;
{
	trace = iop;
	if (trace)
		printf("jtrace  node %d: tracing is on, nodes, master = %d %d\n",node,nodes,master);
	else
		printf("jtrace  node %d: tracing is off\n",node);

	return(0);
}

int jtrace_( iop )
int *iop;
{
	return( jtrace( *iop ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jdcomp - Simple domain decomposition
/*    Allocates a block of data for each processor
/*     
/*  SYNOPSIS:
/*
/*    integer n, jntpn, jnlive
/*    call jdcomp( n, jntpn, jnlive )
/*
/*  DESCRIPTION:
/*    Divides 'n' by number of processors, and allocates 'jntpn'
/*    values to nodes 0 to n-2, and 'jnlive' values to node n-1.
/*    If n is an even multiple of the number of processors, then
/*    jnlive = jntpn.
/*
/*--------------------------------------------------------------------*/

int jdcomp( n, jntpn, jnlive )
int n, *jntpn, *jnlive;
{
	int nrem;

/* Compute number of traces to place on last processor */
	nrem = n % nodes;
	if (nrem > 0) {
		ntpn = (n/nodes) + 1;
		if (node == nodes-1)
			nlive = n - (ntpn*(nodes-1));
		else
			nlive = ntpn;
	}
	else {
		ntpn = n/nodes;
		nlive = ntpn;
	}

	nglob = n;
	*jntpn = ntpn;
	*jnlive = nlive;

	if (trace)
		printf("jdcomp  node %d: nglob, ntpn, nlive = %d %d %d\n",
			node,nglob,ntpn,nlive );

	return(0);
}

/* Fortran interface */

int jdcomp_( n, jntpn, jnlive )
int *n, *jntpn, *jnlive;
{
	return( jdcomp( *n, jntpn, jnlive ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jrange - simple index range computation
/*     
/*  SYNOPSIS:
/*    integer n, j1, jn
/*    call jrange( n, j1, jn )
/*
/*  DESCRIPTION:
/*    Allocates 'n' values across processors using 'jdcomp', then
/*    returns the global index range of the values
/*    allocated to the current processor.
/*
/*--------------------------------------------------------------------*/

int jrange( n, j1, jn )
int n, *j1, *jn;
{

/* Set decomposition if it's not done yet */
	if (n != nglob) jdcomp( n, &ntpn, &nlive );

/* Set range on this processor */
	*j1 = ntpn*node + 1;
	*jn = *j1 + nlive - 1;

	if (trace)
		printf("jrange node %d: n,j1,jn = %d %d %d\n",node,n,*j1,*jn);

	return(0);
}

/* Fortran interface */

int jrange_( n, j1, jn )
int *n, *j1, *jn;
{
	return( jrange( *n, j1, jn ) );
}

/*--------------------------------------------------------------------*/
/*  NAME:
/*    jtran23 - Distributed matrix transpose primitive
/*     
/*  SYNOPSIS:
/*    integer n, a, buf
/*    call jtran23( n, a, buf )
/*
/*  DESCRIPTION:
/*    Transposes last two dimensions of an (n,p,p) matrix, where 'p'
/*    is the number of processors, and the last dimension is 
/*    allocated across processors as in 'jdcomp'.  Uses a standard 
/*    tiling algorithm, developed by Intel SSD. See 'trans.f'
/*    for examples of how this primitive can be used to apply more
/*    general transpose operations.
/*
/*--------------------------------------------------------------------*/

/* This version uses message passing to exchange tiles of data */

int jtran23( n, a, buf )
int n, *a, *buf;
{
#	define MSG 10000
	int myswap, i, len, p, pmax;
	int *ap;

	if (nodes <= 1) return(0);

/* Length in bytes of 1 row (tile) of array */
	len = n * sizeof(int);
 
/* Find power of 2 => number of nodes */
	pmax = 1;
	while (pmax < nodes)
		pmax = 2*pmax;

/* Loop and exchange tiles */
	for (i=1; i<pmax; i++) {
/* Processor to swap with is XOR of processor number and i */
		myswap = node ^ i;
/* Restrict swap in case nodes is not a power of 2 */
		if (myswap < nodes) {
/* tile number to send is same as target processor number */
			ap = a + n*myswap;
/* Swap tiles */
			jsend(MSG+i, ap, len, myswap);
			jrecv(MSG+i, buf, len);
			memcpy( ap, buf, len );
		}
	}

	return(0);
}

/* Fortran interface */

int jtran23_( n, a, buf )
int *n, *a, *buf;
{
	return( jtran23( *n, a, buf ) );
}
